import logging

from enum import Enum, auto
from functools import wraps

from .adapter import get_adapter

_logger = logging.getLogger("transaction")


class Transaction:  # pragma: no cover
    """Abstract base class for Datastore transactions.
    """

    class Propagation(Enum):
        """An enum of the various modes transactions can be run in.
        """

        #: Nested transactions should be grouped together into a single
        #: transaction.
        Nested = auto()

        #: Nested transcations should be run independently of their
        #: parent transactions.
        Independent = auto()

    def begin(self):
        "Start this transaction."
        raise NotImplementedError

    def commit(self):
        "Commit this Transaction to Datastore."
        raise NotImplementedError

    def rollback(self):
        "Roll this Transaction back."
        raise NotImplementedError

    def end(self):
        "Clean up this Transaction object."
        raise NotImplementedError


class TransactionError(Exception):
    """Base class for Transaction errors.
    """


class TransactionFailed(TransactionError):
    """Raised by Adapters when a Transaction cannot be applied.

    Parameters:
      message(str): A message.
      cause(Exception or None): The exception that caused this
        Transaction to fail.
    """

    def __init__(self, message, cause=None):
        self.message = message
        self.cause = cause

    def __str__(self):  # pragma: no cover
        return self.message


class RetriesExceeded(TransactionError):
    """Raised by the transactional decorator when it runs out of
    retries while trying to apply a transaction.

    Parameters:
      cause(TransactionError): The last transaction error that caused
        a retry.
    """

    def __init__(self, cause):
        self.cause = cause

    def __str__(self):  # pragma: no cover
        return str(self.cause)


def transactional(*, adapter=None, retries=3, propagation=Transaction.Propagation.Nested):
    """Decorates functions so that all of their operations (except for
    queries) run inside a Datastore transaction.

    Parameters:
      adapter(Adapter, optional): The Adapter to use when running the
        transaction.  Defaults to the current adapter.
      retries(int, optional): The number of times to retry the
        transaction if it couldn't be committed.
      propagation(Transaction.Propagation, optional): The propagation
        strategy to use. By default, transactions are nested, but you
        can force certain transactions to always run independently.

    Raises:
      anom.RetriesExceeded: When the decorator runbs out of retries
        while trying to commit the transaction.

    Returns:
      callable: The decorated function.
    """
    def decorator(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            nonlocal adapter
            adapter = adapter or get_adapter()
            attempts, cause = 0, None
            while attempts <= retries:
                attempts += 1
                transaction = adapter.transaction(propagation)

                try:
                    transaction.begin()
                    res = fn(*args, **kwargs)
                    transaction.commit()
                    return res

                except TransactionFailed as e:
                    cause = e
                    continue

                except Exception as e:
                    transaction.rollback()
                    raise e

                finally:
                    transaction.end()

            raise RetriesExceeded(cause)
        return inner
    return decorator



import logging
import os
import re
import signal
import shlex
import subprocess

from queue import Empty, Queue
from threading import Thread


#: The command to run in order to start the emulator.
_emulator_command = "gcloud beta emulators datastore start --consistency={consistency:0.2f} --host-port={host}:{port} --no-store-on-disk"  # noqa

#: The regexp that is used to search for env vars in the emulator output.
_env_var_re = re.compile(r"export ([^=]+)=(.+)")

#: The string that is used to determine when the Emulator has finished starting up.
_log_marker = "Dev App Server is now running"


class Emulator:
    """Runs the Cloud Datastore emulator in a subprocess for testing purposes.

    Parameters:
      host(str): The host name the emulator should bind to.
      port(int): The port on which the emulator should listen on.
      consistency(float): A value between 0.0 and 1.0 representing the
        percentage of datastore requests that should succeed.

    Example::

      from anom.testing import Emulator

      @pytest.fixture(scope="session")
      def emulator():
        emulator = Emulator()
        emulator.start(inject=True)
        yield
        emulator.stop()
    """

    def __init__(self, *, host="127.0.0.1", port=9898, consistency=1):
        self._emulator_command = shlex.split(_emulator_command.format(
            host=host, port=port,
            consistency=consistency
        ))

        self._logger = logging.getLogger("Emulator")
        self._proc = None
        self._queue = Queue()
        self._thread = Thread(target=self._run, daemon=True)

    def start(self, *, timeout=15, inject=False):
        """Start the emulator process and wait for it to initialize.

        Parameters:
          timeout(int): The maximum number of seconds to wait for the
            Emulator to start up.
          inject(bool): Whether or not to inject the emulator env vars
            into the current process.

        Returns:
          dict: A dictionary of env vars that can be used to access
          the Datastore emulator.
        """
        try:
            self._thread.start()

            env_vars = self._queue.get(block=True, timeout=timeout)
            if inject:
                os.environ.update(env_vars)

            return env_vars
        except Empty:  # pragma: no cover
            raise RuntimeError("Timed out while waiting for Emulator to start up.")

    def stop(self):
        """Stop the emulator process.

        Returns:
          int: The process return code or None if the process isn't
          currently running.
        """
        if self._proc is not None:
            if self._proc.poll() is None:
                try:
                    os.killpg(self._proc.pid, signal.SIGTERM)
                    _, returncode = os.waitpid(self._proc.pid, 0)
                    self._logger.debug("Emulator process exited with code %d.", returncode)
                    return returncode
                except ChildProcessError:  # pragma: no cover
                    return self._proc.returncode
            return self._proc.returncode  # pragma: no cover
        return None  # pragma: no cover

    def _run(self):
        self._proc = subprocess.Popen(
            self._emulator_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            preexec_fn=os.setsid,
        )

        env_vars = {}
        while self._proc.poll() is None:
            line = self._proc.stdout.readline().strip().decode("utf-8")
            self._logger.debug(line)

            match = _env_var_re.search(line)
            if match:
                name, value = match.groups()
                env_vars[name] = value

            # If no env vars were found this will eventually cause
            # `start` to time out which is what we want since running
            # tests w/o the env vars set up could prove dangerous.
            if _log_marker in line and env_vars:
                self._queue.put(env_vars)



import pylibmc
import uuid

from contextlib import contextmanager
from hashlib import md5
from threading import local

from .. import Adapter, Transaction
from ..properties import Msgpack


class _MemcacheOuterTransaction(Transaction):
    def __init__(self, adapter, ds_transaction):
        self.adapter = adapter
        self.ds_transaction = ds_transaction

        self.batch = []
        self.begin = self.ds_transaction.begin
        self.rollback = self.ds_transaction.rollback

    def _push_keys(self, keys):
        self.batch.extend(keys)

    def commit(self):
        with self.adapter._bust(self.batch):
            self.ds_transaction.commit()

    def end(self):
        self.ds_transaction.end()
        self.adapter._transactions.remove(self)


class _MemcacheInnerTransaction(Transaction):
    def __init__(self, parent, ds_transaction):
        self.parent = parent
        self.ds_transaction = ds_transaction

        self.begin = ds_transaction.begin
        self.commit = ds_transaction.commit
        self.rollback = ds_transaction.rollback

    def end(self):
        self.ds_transaction.end()
        self.adapter._transactions.remove(self)

    def __getattr__(self, name):
        return getattr(self.parent, name)


class MemcacheAdapter(Adapter):
    """Transparently adds memcached-based strongly-consistent caching
    on top of another adapter for delete, get and put operations.

    Parameters:
      client(pylibmc.Client): The memcached client instance to use.
        This is automatically wrapped inside a ThreadMappedPool.
      adapter(Adapter): The adapter to wrap.
      prefix(str, optional): The string keys should be prefixed
        with.  Defaults to ``anom``.
    """

    _state = local()

    _lock_prefix = b"LOCK@"
    _lock_timeout = 60  # seconds
    _item_timeout = 86400  # one day in seconds

    def __init__(self, client, adapter, *, prefix="anom"):
        self.client_pool = pylibmc.ThreadMappedPool(client)
        self.adapter = adapter
        self.prefix = prefix

        self.query = self.adapter.query

    @property
    def _transactions(self):
        "list[Transaction]: The current stack of Transactions."
        transactions = getattr(self._state, "transactions", None)
        if transactions is None:
            transactions = self._state.transactions = []
        return transactions

    def delete_multi(self, keys):
        if self.in_transaction:
            self.current_transaction._push_keys(keys)
            return self.adapter.delete_multi(keys)

        with self._bust(keys):
            return self.adapter.delete_multi(keys)

    def get_multi(self, keys):
        if self.in_transaction:
            return self.adapter.get_multi(keys)

        # Get all the cached keys.
        pairs = {self._convert_key_to_memcache(key): key for key in keys}
        with self.client_pool.reserve() as client:
            mapping = client.get_multi(pairs.keys())

        # Sort out which ones were found in Memcache and which ones we
        # need to get from Datastore.
        found, missing = [None] * len(keys), []
        for memcache_key, anom_key in pairs.items():
            data = mapping.get(memcache_key)
            # If there is no data or the data appears to be locked, we
            # have to grab the entity from datastore.
            if data is None or data.startswith(self._lock_prefix):
                missing.append(anom_key)
                continue

            index = keys.index(anom_key)
            found[index] = Msgpack._loads(data)

        # Get and cache missing keys from Datastore.
        ds_results = self.adapter.get_multi(missing)
        for anom_key, entity in zip(missing, ds_results):
            index = keys.index(anom_key)
            found[index] = entity
            if entity is None:
                continue

            key = self._convert_key_to_memcache(anom_key)
            data = Msgpack._dumps(entity)
            self._cache(key, data)

        return found

    def put_multi(self, requests):
        # Partial keys' cache doesn't need to be cleared since they
        # can't have been already set so we have to collect the full
        # keys and pass them to the transaction.
        full_keys = [request.key for request in requests if not request.key.is_partial]
        if self.in_transaction:
            self.current_transaction._push_keys(full_keys)
            return self.adapter.put_multi(requests)

        with self._bust(full_keys):
            return self.adapter.put_multi(requests)

    def transaction(self, propagation):
        ds_transaction = self.adapter.transaction(propagation)

        if propagation == Transaction.Propagation.Independent:
            transaction = _MemcacheOuterTransaction(self, ds_transaction)
            self._transactions.append(transaction)
            return transaction

        elif propagation == Transaction.Propagation.Nested:
            if self._transactions:
                transaction = _MemcacheInnerTransaction(self.current_transaction, ds_transaction)
            else:
                transaction = _MemcacheOuterTransaction(self, ds_transaction)

            self._transactions.append(transaction)
            return transaction

        else:  # pragma: no cover
            raise ValueError(f"Invalid propagation option {propagation!r}.")

    @property
    def in_transaction(self):
        return bool(self._transactions)

    @property
    def current_transaction(self):
        return self._transactions[-1]

    def _convert_key_to_memcache(self, anom_key):
        digest = md5(str(anom_key).encode("utf-8")).hexdigest()
        return f"{self.prefix}:{digest}"

    @contextmanager
    def _bust(self, keys):
        current_lock = self._lock_value()
        memcache_keys = [self._convert_key_to_memcache(key) for key in keys]
        memcache_pairs = {key: current_lock for key in memcache_keys}

        # Lock the keys so that they can't be set for the duration of
        # the delete (or until timeout).
        with self.client_pool.reserve() as client:
            client.set_multi(memcache_pairs, self._lock_timeout)

        try:
            # Delete the keys from Datastore.
            yield

        finally:
            # Finally, delete the keys from Memcache.
            with self.client_pool.reserve() as client:
                client.delete_multi(memcache_keys)

    def _cache(self, key, data):
        current_lock = self._lock_value()
        with self.client_pool.reserve() as client:
            while True:
                value, cid = client.gets(key)
                # If there is a lock value in Memcache that is
                # different from our own we have to bail.
                if value and value.startswith(self._lock_prefix) and value != current_lock:
                    return

                # If there isn't a value at all, we have to add one
                # and try again.
                if cid is None:
                    client.add(key, current_lock, self._lock_timeout)
                    continue

                try:
                    return client.cas(key, data, cid, self._item_timeout)

                # There is a small chance that between the `gets` and
                # "now" the key will have been deleted by a concurrent
                # process, so we account for that possibility here.
                except pylibmc.NotFound:  # pragma: no cover
                    return

    def _lock_value(self):
        random_value = str(uuid.uuid4()).encode("ascii")
        lock_value = self._lock_prefix + random_value
        return lock_value



from collections import namedtuple

#: The global adapter instance.
_adapter = None


def get_adapter():
    """Get the current global Adapter instance.

    Returns:
      Adapter: The global adapter.  If no global adapter was set, this
      instantiates a :class:`adapters.DatastoreAdapter` and makes it
      the default.
    """
    if _adapter is None:
        from .adapters import DatastoreAdapter
        return set_adapter(DatastoreAdapter())
    return _adapter


def set_adapter(adapter):
    """Set the global Adapter instance.

    Parameters:
      adapter(Adapter): The instance to set as the global adapter.
        Model-specific adapters will not be replaced.

    Returns:
      Adapter: The input adapter.
    """
    global _adapter
    _adapter = adapter
    return _adapter


class PutRequest(namedtuple("PutRequest", ("key", "unindexed", "properties"))):
    """Represents requests to persist individual Models.

    Parameters:
      key(anom.Key): A datastore Key.
      unindexed(list[str]): A list of properties that should not be indexed.
      properties(iter[tuple[str, Property]]): An iterable representing
        the properties that should be stored by this put.
    """


class QueryResponse(namedtuple("QueryResponse", ("entities", "cursor"))):
    """Represents query responses from Datastore.

    Parameters:
      entities(list[tuple[anom.Key, dict]]): The list of results.
      cursor(str): The cursor that points to the next page of results.
        This value must be url-safe.
    """


class Adapter:  # pragma: no cover
    """Abstract base class for Datastore adapters.  Adapters determine
    how your :class:`Models<Model>` interact with the Datastore.
    """

    def delete_multi(self, keys):
        """Delete a list of entities from the Datastore by their
        respective keys.

        Parameters:
          keys(list[anom.Key]): A list of datastore Keys to delete.
        """
        raise NotImplementedError

    def get_multi(self, keys):
        """Get multiple entities from the Datastore by their
        respective keys.

        Parameters:
          keys(list[anom.Key]): A list of datastore Keys to get.

        Returns:
          list[dict]: A list of dictionaries of data that can be loaded
          into individual Models.  Entries for Keys that cannot be
          found are going to be ``None``.
        """
        raise NotImplementedError

    def put_multi(self, requests):
        """Store multiple entities into the Datastore.

        Parameters:
          requests(list[PutRequest]): A list of datastore requets to
            persist a set of entities.

        Returns:
          list[anom.Key]: The list of full keys for each stored
          entity.
        """
        raise NotImplementedError

    def query(self, query, options):
        """Run a query against the datastore.

        Parameters:
          query(Query): The query to run.
          options(QueryOptions): Options that determine how the data
            should be fetched.

        Returns:
          QueryResponse: The query response from Datastore.
        """
        raise NotImplementedError

    def transaction(self, propagation):
        """Create a new Transaction object.

        Parameters:
          propagation(Transaction.Propagation): How the new
            transaction should be propagated with regards to any
            previously-created transactions.

        Returns:
          Transaction: The transaction.
        """
        raise NotImplementedError

    @property
    def in_transaction(self):
        "bool: True if the adapter is currently in a Transaction."
        raise NotImplementedError

    @property
    def current_transaction(self):
        "Transaction: The current Transaction or None."
        raise NotImplementedError



from contextlib import contextmanager
from threading import local

_default_namespace = ""
_namespace = local()


def set_default_namespace(namespace=None):
    """Set the global default namespace.

    Parameters:
      namespace(str): The namespace to set as the global default.

    Returns:
      str: The input namespace.
    """
    global _default_namespace
    _default_namespace = namespace or ""
    return _default_namespace


def get_namespace():
    """str: The namespace for the current thread.
    """
    try:
        return _namespace.current
    except AttributeError:
        return _default_namespace


def set_namespace(namespace=None):
    """Set the current thread-local default namespace.  If namespace
    is None, then the thread-local namespace value is removed, forcing
    `get_namespace()` to return the global default namespace on
    subsequent calls.

    Parameters:
      namespace(str): namespace to set as the current thread-local
        default.

    Returns:
      None
    """
    if namespace is None:
        try:
            del _namespace.current
        except AttributeError:
            pass

    else:
        _namespace.current = namespace


@contextmanager
def namespace(namespace):
    """Context manager for stacking the current thread-local default
    namespace.  Exiting the context sets the thread-local default
    namespace back to the previously-set namespace.  If there is no
    previous namespace, then the thread-local namespace is cleared.

    Example:
      >>> with namespace("foo"):
      ...   with namespace("bar"):
      ...     assert get_namespace() == "bar"
      ...   assert get_namespace() == "foo"

      >>> assert get_namespace() == ""

    Parameters:
      namespace(str): namespace to set as the current thread-local
        default.

    Returns:
      None
    """
    try:
        current_namespace = _namespace.current
    except AttributeError:
        current_namespace = None

    set_namespace(namespace)
    try:
        yield
    finally:
        set_namespace(current_namespace)
